/* ----------------------------------------------------------------------------
 *         ATMEL Microcontroller Software Support 
 * ----------------------------------------------------------------------------
 * Copyright (c) 2008, Atmel Corporation
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer below.
 *
 * Atmel's name may not be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * DISCLAIMER: THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 */
 
//------------------------------------------------------------------------------
//         Headers
//------------------------------------------------------------------------------

#include <board.h>
#include <rtt/rtt.h>
#include <pio/pio.h>
#include <dbgu/dbgu.h>

#include <stdio.h>

#if defined(__ICCARM__)
#include <intrinsics.h>
#endif

//------------------------------------------------------------------------------
//         Local definitions
//------------------------------------------------------------------------------

/// Number of consecutive runs of the testloop() function.
#define NUM_RUNS            25

//------------------------------------------------------------------------------
//         Imported functions
//------------------------------------------------------------------------------

extern void testloop(int);

//------------------------------------------------------------------------------
//         Internal types
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// Describes a possible clock configuration (processor clock & master clock),
/// including the necessary register values.
//------------------------------------------------------------------------------
struct ClockConfiguration {

    /// Processor clock frequency (in MHz).
    unsigned short pck;
    /// Master clock frequency (in MHz).
    unsigned short mck;
    /// CKGR_PLL reqister value.
    unsigned int pllr;
    /// PMC_MCKR register value.
    unsigned int mckr;
};

//------------------------------------------------------------------------------
//         Internal variables
//------------------------------------------------------------------------------

// Clock configurations for the AT91SAM9263-EK
#if defined(at91sam9263)

#define AT91C_CKGR_PLLR     AT91C_CKGR_PLLAR
#define AT91C_PMC_LOCK      AT91C_PMC_LOCKA

static const struct ClockConfiguration clockConfigurations[] = {

    // PCK = 240 MHz, MCK = 120 MHz
    {240, 120, 0x202BC003, 0x00000102},
    // PCK = 120 MHz, MCK = 120 MHz
    {120, 120, 0x20156003, 0x00000002},
    // PCK = 96 MHz, MCK = 48 MHz
    {96, 48, 0x2057400F, 0x00000102},
    // PCK = 48 MHz, MCK = 48 MHz
    {48, 48, 0x2057400F, 0x00000006}
};

// Clock configurations for the AT91SAM9G20-EK
#elif defined(at91sam9g20)

#define AT91C_CKGR_PLLR     AT91C_CKGR_PLLAR
#define AT91C_PMC_LOCK      AT91C_PMC_LOCKA

static const struct ClockConfiguration clockConfigurations[] = {

    // PCK = 400 MHz, MCK = 133 MHz
    {400, 133, 0x202A0101, 0x00001302},
    // PCK = 200 MHz, MCK = 100 MHz
    {200, 100, 0x202A3F01, 0x0000010A},
    // PCK = 96 MHz, MCK = 48 MHz
    {96, 48, 0x207C3F03, 0x0000010E},
    // PCK = 48 MHz, MCK = 48 MHz
    {48, 48, 0x207C3F03, 0x00000012}
};

// Clock configurations for the AT91SAM9XE-EK, AT91SAM9261-EK
#elif defined(at91sam9xe128) || defined(at91sam9xe256) || defined(at91sam9xe512) \
      || defined(at91sam9261)

#define AT91C_CKGR_PLLR     AT91C_CKGR_PLLAR
#define AT91C_PMC_LOCK      AT91C_PMC_LOCKA

static const struct ClockConfiguration clockConfigurations[] = {

    // PCK = 200 MHz, MCK = 100 MHz
    {200, 100, 0x208CBF0D, 0x00000102},
    // PCK = 100 MHz, MCK = 100 MHz
    {100, 100, 0x20273F07, 0x00000002},
    // PCK = 96 MHz, MCK = 48 MHz
    {96, 48, 0x20483F0E, 0x00000102},
    // PCK = 48 MHz, MCK = 48 MHz
    {48, 48, 0x20483F0E, 0x00000006}
};

// Clock configurations for the AT91SAM9RL64-EK, AT91CAP9-DK, AT91CAP9-STK
#elif defined(at91sam9rl64) || defined(at91cap9)

#define AT91C_CKGR_PLLR     AT91C_CKGR_PLLAR
#define AT91C_PMC_LOCK      AT91C_PMC_LOCKA

static const struct ClockConfiguration clockConfigurations[] = {

    // PCK = 200 MHz, MCK = 100 MHz
    {200, 100, 0x2031BF03, 0x00000102},
    // PCK = 100 MHz, MCK = 100 MHz
    {100, 100, 0x20183F03, 0x00000002},
    // PCK = 96 MHz, MCK = 48 MHz
    {96, 48, 0x20073F01, 0x00000102},
    // PCK = 48 MHz, MCK = 48 MHz
    {48, 48, 0x20073F01, 0x00000006}
};

// Clock configuration for the AT91SAM7X-EK, AT91SAM7S-EK, AT91SAM7SE-EK
#elif defined(at91sam7x128) || defined(at91sam7x256) || defined(at91sam7x512) \
      || defined(at91sam7xc128) || defined(at91sam7xc256) || defined(at91sam7xc512) \
      || defined(at91sam7s16) || defined(at91sam7s161) || defined(at91sam7s32) \
      || defined(at91sam7s321) || defined(at91sam7s64) || defined(at91sam7s128) \
      || defined(at91sam7s256) || defined(at91sam7s512) || defined(at91sam7se32) \
      || defined(at91sam7se256) || defined(at91sam7se512)

static const struct ClockConfiguration clockConfigurations[] = {

    // PCK = 55 MHz, MCK = 55 MHz
    {55, 55, 0x006A3F12, 0x00000007},
    // PCK = 48 MHz, MCK = 48 MHz
    {48, 48, 0x00483F0E, 0x00000007},
};

// Clock configuration for the AT91SAM7A3
#elif defined(at91sam7a3)

static const struct ClockConfiguration clockConfigurations[] = {

    // PCK = 60 MHz, MCK = 60 MHz
    {60, 60, 0x000C3F02, 0x00000007},
    // PCK = 48 MHz, MCK = 48 MHz
    {48, 48, 0x00483F0E, 0x00000007}
};

// No clock configuration
#else
    #error No clock configuration for this board.
#endif

/// Number of available clock configurations
const unsigned char numConfigurations = sizeof(clockConfigurations)
                                        / sizeof(struct ClockConfiguration);

/// List of pins to configure.
static const Pin pins[] = {PINS_DBGU};

//------------------------------------------------------------------------------
//         Internal functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// Computes the number of dhrystones per second that the device can produce
/// under the current configuration. Outputs four measures on the DBGU.
//------------------------------------------------------------------------------
static void ComputeDhrystonesPerSecond()
{
    unsigned int startTime;
    unsigned int dhrystonesPerSecond;
    unsigned char i;

    printf("Computing dhrystones per second ...\n\r");

    // Configure the RTT
    RTT_SetPrescaler(AT91C_BASE_RTTC, 32768);

    // Perform measures
    for (i=0; i < 4; i++) {
    
        dhrystonesPerSecond = 0;  
        
        // Wait for the next second
    	startTime = RTT_GetTime(AT91C_BASE_RTTC);
    	while (startTime == RTT_GetTime(AT91C_BASE_RTTC));
    
        // Go through test loop
        startTime = RTT_GetTime(AT91C_BASE_RTTC);
        while (RTT_GetTime(AT91C_BASE_RTTC) == startTime) {
    
            testloop(NUM_RUNS);
            dhrystonesPerSecond += NUM_RUNS;
        }
    
        printf(" - %u dhrystones per second\n\r", dhrystonesPerSecond);
    }
    printf("Finished\n\r");
}

//------------------------------------------------------------------------------
/// Displays the user menu on the DBGU.
//------------------------------------------------------------------------------
static void DisplayMenu()
{
    unsigned int i;

    printf("\n\rMenu :\n\r");
    printf("------\n\r");

    for (i = 0; i < numConfigurations; i++) {

        printf("  %u: Set PCK = %3d MHz, MCK = %3d MHz\n\r",
               i,
               clockConfigurations[i].pck,
               clockConfigurations[i].mck);
    }
#if defined(at91sam9263) || defined(at91sam9rl64) || defined(at91sam9260) \
    || defined(at91sam9xe128) || defined(at91sam9xe256) || defined(at91sam9xe512) \
    || defined(at91sam9g20) || defined(at91sam9261) || defined(at91cap9)
    printf("  i: Toggle instruction cache\n\r");
#endif
    printf("  m: Perform measurements\n\r");
    printf("  h: Display this menu again\n\r\n\r");
}

#if defined(at91sam9263) || defined(at91sam9rl64) || defined(at91sam9260) \
    || defined(at91sam9xe128) || defined(at91sam9xe256) || defined(at91sam9xe512) \
    || defined(at91sam9g20) || defined(at91sam9261) || defined(at91cap9)
//------------------------------------------------------------------------------
/// Reads and returns the value of the CP15 control register.
//------------------------------------------------------------------------------
unsigned int ReadControlRegister(void)
{
    unsigned int control = 0;

#if defined(__GNUC__)
    asm("MRC p15, 0, %0, c1, c0, 0;" : "=r" (control));
#elif defined(__ICCARM__)
    control = __MRC(15, 0, 1, 0, 0);
#endif

    return control;
}

//------------------------------------------------------------------------------
/// Changes the value of the CP15 control register.
/// \param control  New CP15 value.
//------------------------------------------------------------------------------
void WriteControlRegister(unsigned int control)
{
#if defined(__GNUC__)
    asm("MCR p15, 0, %0, c1, c0, 0;" : : "r" (control));
#elif defined(__ICCARM__)
    __MCR(15, 0, control, 1, 0, 0);
#endif
}

//------------------------------------------------------------------------------
/// Toggles the status of the instruction cache (enabled/disabled).
//------------------------------------------------------------------------------
static void ToggleInstructionCache()
{
    unsigned int control;

    control = ReadControlRegister();

    // Check if cache is enabled
    if ((control & (1 << 12)) == 0) {

        control |= (1 << 12);
        printf("Instruction cache enabled.\n\r");
    }
    else {

        control &= ~(1 << 12);
        printf("Instruction cache disabled.\n\r");
    }

    WriteControlRegister(control);
}
#endif 

//------------------------------------------------------------------------------
/// Sets the specified clock configuration.
/// \param configuration  Index of the configuration to set.
//------------------------------------------------------------------------------
static void SetClockConfiguration(unsigned char configuration)
{
    printf("Setting clock configuration #%d ... ", configuration);

    // Switch to main oscillator in two operations
    AT91C_BASE_PMC->PMC_MCKR = (AT91C_BASE_PMC->PMC_MCKR & ~AT91C_PMC_CSS) | AT91C_PMC_CSS_MAIN_CLK;
    while ((AT91C_BASE_PMC->PMC_SR & AT91C_PMC_MCKRDY) == 0);

    // Configure PLL
    *AT91C_CKGR_PLLR = clockConfigurations[configuration].pllr;
    while ((AT91C_BASE_PMC->PMC_SR & AT91C_PMC_LOCK) == 0);

    // Configure master clock in two operations
    AT91C_BASE_PMC->PMC_MCKR = (clockConfigurations[configuration].mckr & ~AT91C_PMC_CSS) | AT91C_PMC_CSS_MAIN_CLK;
    while ((AT91C_BASE_PMC->PMC_SR & AT91C_PMC_MCKRDY) == 0);
    AT91C_BASE_PMC->PMC_MCKR = clockConfigurations[configuration].mckr;
    while ((AT91C_BASE_PMC->PMC_SR & AT91C_PMC_MCKRDY) == 0);

    // DBGU reconfiguration
    DBGU_Configure(DBGU_STANDARD, 115200, clockConfigurations[configuration].mck*1000000);
    printf("done.\n\r");
}

//------------------------------------------------------------------------------
//         Exported functions
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
/// Displays a menu to let the user either change the clock configuration of
/// the device, enable/disable options (i.e. instruction cache) or start a
/// measure.
//------------------------------------------------------------------------------
int main()
{
    unsigned char key;
    
    // DBGU configuration
    PIO_Configure(pins, PIO_LISTSIZE(pins));
    DBGU_Configure(DBGU_STANDARD, 115200, BOARD_MCK);
	printf("-- Basic Dhrystone Project 1.4 --\n\r");

    // Display menu
    DisplayMenu();

    while (1) {

        // Get keypress
        key = DBGU_GetChar();

        // Check key
        //----------
        // Set configuration #n
        if ((key >= '0') && (key <= ('0' + numConfigurations - 1))) {

            SetClockConfiguration(key - '0');
        }
        // Menu
        else if (key == 'h') {

            DisplayMenu();
        }
        // Measurement
        else if (key == 'm') {

            ComputeDhrystonesPerSecond();
        }
#if defined(at91sam9263) || defined(at91sam9rl64) || defined(at91sam9260) \
    || defined(at91sam9xe128) || defined(at91sam9xe256) || defined(at91sam9xe512) \
    || defined(at91sam9g20) || defined(at91sam9261) || defined(at91cap9)
        // Instruction cache
        else if (key == 'i') {

            ToggleInstructionCache();
        }
#endif
        // Unrecognized option
        else {

            printf("No menu entry for '%c'\n\r", key);
        }
    }
}

